@@ -5,8 +5,6 @@ import re
 from collections import OrderedDict, defaultdict
 from functools import reduce
 
-import attr
-
 from . import biblio, config, h, messages as m, refs, t
 from .line import Line
 
@@ -22,9 +20,7 @@ if t.TYPE_CHECKING:
     InfoTreeT: t.TypeAlias = list[defaultdict[str, list[str]]]
 
     class TransformFuncT(t.Protocol):
-        def __call__(
-            self, lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT
-        ) -> list[str]:
+        def __call__(self, lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
             ...
 
 
@@ -81,39 +77,45 @@ def transformDataBlocks(doc: t.SpecT, lines: list[Line] | list[str]) -> list[Lin
         "include-raw": transformIncludeRaw,
         "pre": transformPre,
     }
-    blockType = ""
+    blockType: str = ""
     tagName = ""
+    blockTag: h.StartTag | None = None
     blockLines: list[Line] = []
     newLines: list[Line] = []
     for line in _lines:
         # Look for the start of a block.
-        match = re.match(r"\s*<(pre|xmp)[\s>]", line.text, re.I)
+        startTag, _ = h.parseStartTag(h.Stream(line.text.lstrip(), startLine=line.i), 0)
+        if startTag is h.Failure:
+            startTag = None
+        if startTag and startTag.tag not in ("pre", "xmp"):
+            startTag = None
+
         # Note that, by design, I don't pay attention to anything on the same line as the start tag,
         # unless it's single-line.
-        if match and not inBlock:
+        if startTag and not inBlock:
+            blockTag = startTag.finalize()
+            assert blockTag is not None
+            blockTag.line = line.i
             inBlock = True
-            tagName = match.group(1)
-            blockClasses = classesFromLine(line)
-            seenClasses: list[str] = []
-            for blockType in blockTypes:
-                if blockType in blockClasses:
-                    seenClasses.append(blockType)
-            if not seenClasses:
-                blockType = "pre"
-            elif len(seenClasses) == 1:
-                blockType = seenClasses[0]
-            else:
+            tagName = blockTag.tag
+            foundTypes = blockTypeFromTag(blockTag, list(blockTypes.keys()))
+            if len(foundTypes) > 1:
+                typeList = config.englishFromList((f"'{x}'" for x in foundTypes), "and")
                 m.die(
-                    "Found {} classes on the <{}>, so can't tell which to process the block as. Please use only one.".format(
-                        config.englishFromList((f"'{x}'" for x in seenClasses), "and"), tagName
-                    ),
+                    f"Found {typeList} classes on the <{blockTag.tag}>, so can't tell which to process the block as. Please use only one.",
                     lineNum=line.i,
                 )
                 blockType = "pre"
+            elif len(foundTypes) == 0:
+                blockType = "pre"
+            else:
+                blockType = foundTypes[0]
+            assert blockType is not None
         # Look for the end of a block.
         match = re.match(r"(.*)</" + tagName + ">(.*)", line.text, re.I)
         if match and inBlock:
             inBlock = False
+            assert blockTag is not None
             if len(blockLines) == 0:
                 # Single-line <pre>.
                 match = re.match(r"(\s*<{0}[^>]*>)(.*)</{0}>(.*)".format(tagName), line.text, re.I)
@@ -123,9 +125,8 @@ def transformDataBlocks(doc: t.SpecT, lines: list[Line] | list[str]) -> list[Lin
                     continue
                 repl = blockTypes[blockType](
                     lines=[match.group(2)],
-                    tagName=tagName,
+                    startTag=blockTag,
                     firstLine=match.group(1),
-                    lineNum=line.i,
                     doc=doc,
                 )
                 newLines.extend(Line(line.i, x) for x in repl)
@@ -136,9 +137,8 @@ def transformDataBlocks(doc: t.SpecT, lines: list[Line] | list[str]) -> list[Lin
                 # Remove the tag from the line.
                 repl = blockTypes[blockType](
                     lines=cleanPrefix([x.text for x in blockLines[1:]]),
-                    tagName=tagName,
+                    startTag=blockTag,
                     firstLine=blockLines[0].text,
-                    lineNum=blockLines[0].i,
                     doc=doc,
                 )
                 newLines.extend(Line(blockLines[0].i, x) for x in repl)
@@ -149,9 +149,8 @@ def transformDataBlocks(doc: t.SpecT, lines: list[Line] | list[str]) -> list[Lin
                 # Process the stuff before it, preserve the stuff after it.
                 repl = blockTypes[blockType](
                     lines=cleanPrefix([x.text for x in blockLines[1:]] + [match.group(1)]),
-                    tagName=tagName,
+                    startTag=blockTag,
                     firstLine=blockLines[0].text,
-                    lineNum=blockLines[0].i,
                     doc=doc,
                 )
                 newLines.extend(Line(blockLines[0].i, x) for x in repl)
@@ -159,6 +158,7 @@ def transformDataBlocks(doc: t.SpecT, lines: list[Line] | list[str]) -> list[Lin
                 newLines.append(line)
             tagName = ""
             blockType = ""
+            blockTag = None
             blockLines = []
             continue
         if inBlock:
@@ -201,19 +201,25 @@ def getWsPrefix(line: str) -> str:
     return t.cast(str, match.group(1))
 
 
-def transformPre(lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT) -> list[str]:
+def blockTypeFromTag(tag: h.StartTag, blockTypes: t.Sequence[str]) -> list[str]:
+    # See which of the designated blockTypes classes
+    # are present in the tag's classes.
+    return list(tag.classes & set(blockTypes))
+
+
+def transformPre(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
     # If the last line in the source is a </code></pre>,
     # the generic processor will turn that into a final </code> line,
     # which'll mess up the indent finding.
     # Instead, specially handle this case.
     if len(lines) == 0:
-        return [firstLine, f"</{tagName}>"]
+        return [firstLine, f"</{startTag.tag}>"]
 
     if re.match(r"\s*</code>\s*$", lines[-1]):
-        lastLine = f"</code></{tagName}>"
+        lastLine = f"</code></{startTag.tag}>"
         lines = lines[:-1]
     else:
-        lastLine = f"</{tagName}>"
+        lastLine = f"</{startTag.tag}>"
 
     if len(lines) == 0:
         return [firstLine, lastLine]
@@ -245,12 +251,12 @@ def transformPre(lines: list[str], tagName: str, firstLine: str, lineNum: int |
     return lines
 
 
-def transformSimpleDef(lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT) -> list[str]:
-    rows = parseDefBlock(lines, "simpledef")
-    lineNumAttr = ""
-    if lineNum is not None:
-        lineNumAttr = f" line-number={lineNum}"
-    ret = [f"<table class='def'{lineNumAttr}>"]
+def transformSimpleDef(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
+    rows = parseDefBlock(lines, "simpledef", lineNum=startTag.line)
+    newStartTag = startTag.clone(tag="table")
+    newStartTag.classes.remove("simpledef")
+    newStartTag.classes.add("def")
+    ret = [str(newStartTag)]
     for key, val in rows.items():
         ret.append(f"<tr><th>{key}<td>{val}")
     ret.append("</table>")
@@ -261,26 +267,23 @@ def transformSimpleDef(lines: list[str], tagName: str, firstLine: str, lineNum:
     return ret
 
 
-def transformPropdef(lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT) -> list[str]:
+def transformPropdef(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
     attrs: OrderedDict[str, str | None] = OrderedDict()
-    parsedAttrs = parseDefBlock(lines, "propdef")
+    parsedAttrs = parseDefBlock(lines, "propdef", lineNum=startTag.line)
     # Displays entries in the order specified in attrs,
     # then if there are any unknown parsedAttrs values,
     # they're displayed afterward in the order they were specified.
     # attrs with a value of None are required to be present in parsedAttrs;
     # attrs with any other value are optional, and use the specified value if not present in parsedAttrs
-    forHint = ""
+    newStartTag = startTag.clone(tag="table")
+    newStartTag.classes.add("def")
+    newStartTag.classes.add("propdef")
     if "Name" in parsedAttrs:
-        forHint = " data-link-for-hint='{}'".format(parsedAttrs["Name"].split(",")[0].strip())
-    lineNumAttr = ""
-    if lineNum is not None:
-        lineNumAttr = f" line-number={lineNum}"
+        newStartTag.attrs["data-link-for-hint"] = parsedAttrs["Name"].split(",")[0].strip()
     if "partial" in firstLine or "New values" in parsedAttrs:
         attrs["Name"] = None
         attrs["New values"] = None
-        ret = [
-            "<table class='def propdef partial'{forHint}{lineNumAttr}>".format(forHint=forHint, lineNumAttr=lineNumAttr)
-        ]
+        newStartTag.classes.add("partial")
     elif "shorthand" in firstLine:
         attrs["Name"] = None
         attrs["Value"] = None
@@ -294,7 +297,6 @@ def transformPropdef(lines: list[str], tagName: str, firstLine: str, lineNum: in
         ]:
             attrs[defaultKey] = "see individual properties"
         attrs["Canonical order"] = "per grammar"
-        ret = ["<table class='def propdef'{forHint}{lineNumAttr}>".format(forHint=forHint, lineNumAttr=lineNumAttr)]
     else:
         attrs["Name"] = None
         attrs["Value"] = None
@@ -305,7 +307,6 @@ def transformPropdef(lines: list[str], tagName: str, firstLine: str, lineNum: in
         attrs["Computed value"] = "as specified"
         attrs["Canonical order"] = "per grammar"
         attrs["Animation type"] = None
-        ret = ["<table class='def propdef'{forHint}{lineNumAttr}>".format(forHint=forHint, lineNumAttr=lineNumAttr)]
     # We are in the process of migrating specs from using 'Animatable' to
     # using 'Animation type'. If we find 'Animatable' in the parsed attributes,
     # drop the default 'Animation type' entry.
@@ -319,7 +320,9 @@ def transformPropdef(lines: list[str], tagName: str, firstLine: str, lineNum: in
             val = parsedAttrs[key]
         elif val is None:
             # Required key, not provided
-            m.die(f"The propdef for '{parsedAttrs.get('Name', '???')}' is missing a '{key}' line.", lineNum=lineNum)
+            m.die(
+                f"The propdef for '{parsedAttrs.get('Name', '???')}' is missing a '{key}' line.", lineNum=startTag.line
+            )
             continue
         else:
             # Optional key, just use default
@@ -330,6 +333,7 @@ def transformPropdef(lines: list[str], tagName: str, firstLine: str, lineNum: in
         if key not in attrs:
             attrsToPrint.append((key, val))
 
+    ret = [str(newStartTag)]
     for key, val in attrsToPrint:
         tr = "<tr>"
         th = f"<th>{key}:"
@@ -369,32 +373,21 @@ def transformPropdef(lines: list[str], tagName: str, firstLine: str, lineNum: in
 # TODO: Make these functions match transformPropdef's new structure
 
 
-def transformDescdef(lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT) -> list[str]:
-    lineNumAttr = ""
-    if lineNum is not None:
-        lineNumAttr = f" line-number={lineNum}"
-    vals = parseDefBlock(lines, "descdef")
+def transformDescdef(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
+    newStartTag = startTag.clone(tag="table")
+    newStartTag.classes.add("def")
+    newStartTag.classes.add("descdef")
+    vals = parseDefBlock(lines, "descdef", lineNum=startTag.line)
+    newStartTag.attrs["data-dfn-for"] = vals.get("For", "")
     if "partial" in firstLine or "New values" in vals:
         requiredKeys = ["Name", "For"]
-        ret = [
-            "<table class='def descdef partial' data-dfn-for='{0}'{lineNumAttr}>".format(
-                vals.get("For", ""), lineNumAttr=lineNumAttr
-            )
-        ]
+        newStartTag.classes.add("partial")
     if "mq" in firstLine:
         requiredKeys = ["Name", "For", "Value"]
-        ret = [
-            "<table class='def descdef mq' data-dfn-for='{0}'{lineNumAttr}>".format(
-                vals.get("For", ""), lineNumAttr=lineNumAttr
-            )
-        ]
+        newStartTag.classes.add("mq")
     else:
         requiredKeys = ["Name", "For", "Value", "Initial"]
-        ret = [
-            "<table class='def descdef' data-dfn-for='{0}'{lineNumAttr}>".format(
-                vals.get("For", ""), lineNumAttr=lineNumAttr
-            )
-        ]
+    ret = [str(newStartTag)]
     for key in requiredKeys:
         if key == "For":
             ret.append("<tr><th>{}:<td><a at-rule>{}</a>".format(key, vals.get(key, "")))
@@ -403,7 +396,7 @@ def transformDescdef(lines: list[str], tagName: str, firstLine: str, lineNum: in
         elif key in vals:
             ret.append("<tr><th>{}:<td>{}".format(key, vals.get(key, "")))
         else:
-            m.die(f"The descdef for '{vals.get('Name', '???')}' is missing a '{key}' line.", lineNum=lineNum)
+            m.die(f"The descdef for '{vals.get('Name', '???')}' is missing a '{key}' line.", lineNum=startTag.line)
             continue
     for key, val in vals.items():
         if key in requiredKeys:
@@ -417,25 +410,24 @@ def transformDescdef(lines: list[str], tagName: str, firstLine: str, lineNum: in
     return ret
 
 
-def transformElementdef(lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT) -> list[str]:
-    lineNumAttr = ""
-    if lineNum is not None:
-        lineNumAttr = f" line-number={lineNum}"
+def transformElementdef(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
+    newStartTag = startTag.clone(tag="table")
+    newStartTag.classes.add("def")
+    newStartTag.classes.add("elementdef")
     attrs: OrderedDict[str, str | None] = OrderedDict()
-    parsedAttrs = parseDefBlock(lines, "elementdef")
+    parsedAttrs = parseDefBlock(lines, "elementdef", lineNum=startTag.line)
     if "Attribute groups" in parsedAttrs or "Attributes" in parsedAttrs:
         html = "<ul>"
         if "Attribute groups" in parsedAttrs:
             groups = [x.strip() for x in parsedAttrs["Attribute groups"].split(",")]
             for group in groups:
-                html += "<li><a dfn data-element-attr-group{lineNumAttr}>{0}</a>".format(group, lineNumAttr=lineNumAttr)
+                html += f"<li><a dfn data-element-attr-group line-number={startTag.line}>{group}</a>"
             del parsedAttrs["Attribute groups"]
         if "Attributes" in parsedAttrs:
             atts = [x.strip() for x in parsedAttrs["Attributes"].split(",")]
             for att in atts:
-                html += "<li><a element-attr for='{1}'{lineNumAttr}>{0}</a>".format(
-                    att, parsedAttrs.get("Name", ""), lineNumAttr=lineNumAttr
-                )
+                forVal = parsedAttrs.get("Name", "")
+                html += f"<li><a element-attr for='{forVal}' line-number={startTag.line}>{att}</a>"
         html += "</ul>"
         parsedAttrs["Attributes"] = html
 
@@ -450,7 +442,7 @@ def transformElementdef(lines: list[str], tagName: str, firstLine: str, lineNum:
     attrs["Content model"] = None
     attrs["Attributes"] = None
     attrs["Dom interfaces"] = None
-    ret = [f"<table class='def elementdef'{lineNumAttr}>"]
+    ret = [str(newStartTag)]
     for key, val in attrs.items():
         if key in parsedAttrs or val is not None:
             if key in parsedAttrs:
@@ -459,33 +451,26 @@ def transformElementdef(lines: list[str], tagName: str, firstLine: str, lineNum:
             if key == "Name":
                 ret.append("<tr><th>Name:<td>")
                 ret.append(
-                    ", ".join(
-                        "<dfn element{lineNumAttr}>{0}</dfn>".format(x.strip(), lineNumAttr=lineNumAttr)
-                        for x in val.split(",")
-                    )
+                    ", ".join(f"<dfn element line-number={startTag.line}>{x.strip()}</dfn>" for x in val.split(","))
                 )
             elif key == "Content model":
                 ret.append(f"<tr><th>{key}:<td>")
                 ret.extend(val.split("\n"))
             elif key == "Categories":
                 ret.append("<tr><th>Categories:<td>")
-                ret.append(
-                    ", ".join(
-                        "<a dfn{lineNumAttr}>{0}</a>".format(x.strip(), lineNumAttr=lineNumAttr) for x in val.split(",")
-                    )
-                )
+                ret.append(", ".join(f"<a dfn line-number={startTag.line}>{x.strip()}</a>" for x in val.split(",")))
             elif key == "Dom interfaces":
                 ret.append("<tr><th>DOM Interfaces:<td>")
                 ret.append(
-                    ", ".join(
-                        "<a interface{lineNumAttr}>{0}</a>".format(x.strip(), lineNumAttr=lineNumAttr)
-                        for x in val.split(",")
-                    )
+                    ", ".join(f"<a interface line-number={startTag.line}>{x.strip()}</a>" for x in val.split(","))
                 )
             else:
                 ret.append(f"<tr><th>{key}:<td>{val}")
         else:
-            m.die(f"The elementdef for '{parsedAttrs.get('Name', '???')}' is missing a '{key}' line.", lineNum=lineNum)
+            m.die(
+                f"The elementdef for '{parsedAttrs.get('Name', '???')}' is missing a '{key}' line.",
+                lineNum=startTag.line,
+            )
             continue
     for key, val in parsedAttrs.items():
         if key in attrs:
@@ -499,32 +484,31 @@ def transformElementdef(lines: list[str], tagName: str, firstLine: str, lineNum:
     return ret
 
 
-def transformArgumentdef(
-    lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT
-) -> list[str]:
-    lineNumAttr = ""
-    if lineNum is not None:
-        lineNumAttr = f" line-number={lineNum}"
-    attrs = parseDefBlock(lines, "argumentdef", capitalizeKeys=False, lineNum=lineNum)
-    el = h.parseHTML(firstLine + "</pre>")[0]
-    if "for" in el.attrib:
-        forValue = t.cast(str, el.get("for"))
-        el.set("data-dfn-for", forValue)
+def transformArgumentdef(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
+    newStartTag = startTag.clone(tag="table")
+    newStartTag.classes.add("def")
+    newStartTag.classes.add("argumentdef")
+    attrs = parseDefBlock(lines, "argumentdef", capitalizeKeys=False, lineNum=startTag.line)
+    if "for" in newStartTag.attrs:
+        forValue = newStartTag.attrs["for"]
         if "/" in forValue:
             interface, method = forValue.split("/")
         else:
-            m.die(f"Argumentdef for='' values need to specify interface/method(). Got '{forValue}'.", lineNum=lineNum)
+            m.die(
+                f"Argumentdef for='' values need to specify interface/method(). Got '{forValue}'.",
+                lineNum=startTag.line,
+            )
             return []
-        h.removeAttr(el, "for")
+        newStartTag.attrs["data-dfn-for"] = newStartTag.attrs["for"]
+        del newStartTag.attrs["for"]
     else:
-        m.die("Argumentdef blocks need a for='' attribute specifying their method.", lineNum=lineNum)
+        m.die("Argumentdef blocks need a for='' attribute specifying their method.", lineNum=startTag.line)
         return []
-    h.addClass(doc, el, "data")
-    rootAttrs = " ".join(f"{str(k)}='{h.escapeAttr(str(v))}'" for k, v in el.attrib.items())
+    newStartTag.classes.add("data")
     text = (
-        """
-<table {attrs}{lineNumAttr}>
-<caption>Arguments for the <a idl lt='{method}' for='{interface}'{lineNumAttr}>{interface}.{method}</a> method.</caption>
+        f"""
+{newStartTag}
+<caption>Arguments for the <a idl lt='{method}' for='{interface}' line-number={startTag.line}>{interface}.{method}</a> method.</caption>
 <thead>
 <tr>
 <th>Parameter
@@ -532,20 +516,16 @@ def transformArgumentdef(
 <th style="text-align:center">Nullable
 <th style="text-align:center">Optional
 <th>Description
-<tbody>""".format(
-            attrs=rootAttrs, interface=interface, method=method, lineNumAttr=lineNumAttr
-        )
+<tbody>"""
         + "\n".join(
             [
-                """
+                f"""
 <tr>
-<td><dfn argument{lineNumAttr}>{0}</dfn>
+<td><dfn argument line-number={startTag.line}>{param}</dfn>
 <td>
 <td style="text-align:center">
 <td style="text-align:center">
-<td>{1}""".format(
-                    param, desc, lineNumAttr=lineNumAttr
-                )
+<td>{desc}"""
                 for param, desc in attrs.items()
             ]
         )
@@ -586,12 +566,15 @@ def parseDefBlock(
     return vals
 
 
-def transformRailroad(lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT) -> list[str]:
+def transformRailroad(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
     import io
 
     from . import railroadparser
 
-    ret = ["<div class='railroad'>"]
+    newStartTag = startTag.clone(tag="div")
+    newStartTag.classes.add("railroad")
+
+    ret = [str(newStartTag)]
     doc.extraStyles[
         "style-railroad"
     ] = """
@@ -650,7 +633,7 @@ def transformRailroad(lines: list[str], tagName: str, firstLine: str, lineNum: i
     return []
 
 
-def transformBiblio(lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT) -> list[str]:
+def transformBiblio(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
     storage: t.BiblioStorageT = defaultdict(list)
     biblio.processSpecrefBiblioFile("".join(lines), storage, order=1)
     for k, vs in storage.items():
@@ -659,9 +642,9 @@ def transformBiblio(lines: list[str], tagName: str, firstLine: str, lineNum: int
     return []
 
 
-def transformAnchors(lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT) -> list[str]:
-    anchors = parseInfoTree(lines, doc.md.indent, lineNum)
-    processAnchors(anchors, doc, lineNum)
+def transformAnchors(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
+    anchors = parseInfoTree(lines, doc.md.indent, lineNum=startTag.line)
+    processAnchors(anchors, doc, lineNum=startTag.line)
     return []
 
 
@@ -751,11 +734,9 @@ def processAnchors(anchors: InfoTreeT, doc: t.SpecT, lineNum: int | None = None)
             doc.refs.anchorBlockRefs.addMethodVariants(anchor["text"][0], anchor.get("for", []), doc.md.shortname)
 
 
-def transformLinkDefaults(
-    lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT
-) -> list[str]:
-    lds = parseInfoTree(lines, doc.md.indent, lineNum)
-    processLinkDefaults(lds, doc, lineNum)
+def transformLinkDefaults(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
+    lds = parseInfoTree(lines, doc.md.indent, lineNum=startTag.line)
+    processLinkDefaults(lds, doc, lineNum=startTag.line)
     return []
 
 
@@ -790,11 +771,9 @@ def processLinkDefaults(lds: InfoTreeT, doc: t.SpecT, lineNum: int | None = None
             doc.md.linkDefaults[text].append((spec, type, status, None))
 
 
-def transformIgnoredSpecs(
-    lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT
-) -> list[str]:
-    specs = parseInfoTree(lines, doc.md.indent, lineNum)
-    processIgnoredSpecs(specs, doc, lineNum)
+def transformIgnoredSpecs(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
+    specs = parseInfoTree(lines, doc.md.indent, lineNum=startTag.line)
+    processIgnoredSpecs(specs, doc, lineNum=startTag.line)
     return []
 
 
@@ -821,12 +800,12 @@ def processIgnoredSpecs(specs: InfoTreeT, doc: t.SpecT, lineNum: int | None = No
                 doc.refs.ignoredSpecs.add(specName)
 
 
-def transformInfo(lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT) -> list[str]:
+def transformInfo(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
     # More generic InfoTree system.
     # A <pre class=info> can contain any of the InfoTree collections,
     # identified by an 'info' line.
-    infos = parseInfoTree(lines, doc.md.indent, lineNum)
-    processInfo(infos, doc, lineNum)
+    infos = parseInfoTree(lines, doc.md.indent, lineNum=startTag.line)
+    processInfo(infos, doc, lineNum=startTag.line)
     return []
 
 
@@ -852,11 +831,10 @@ def processInfo(infos: InfoTreeT, doc: t.SpecT, lineNum: int | None = None) -> N
         knownInfoTypes[infoType](infoItem, doc, lineNum=0)
 
 
-def transformInclude(lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT) -> list[str]:
-    lineNumAttr = ""
-    if lineNum is not None:
-        lineNumAttr = f" line-number={lineNum}"
-    infos = parseInfoTree(lines, doc.md.indent, lineNum)
+def transformInclude(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
+    newStartTag = startTag.clone(tag="pre")
+    newStartTag.classes.add("include")
+    infos = parseInfoTree(lines, doc.md.indent, lineNum=startTag.line)
     path = None
     macros = {}
     for info in infos:
@@ -864,7 +842,7 @@ def transformInclude(lines: list[str], tagName: str, firstLine: str, lineNum: in
             if path is None:
                 path = info["path"][0]
             else:
-                m.die("Include blocks must only contain a single 'path'.", lineNum=lineNum)
+                m.die("Include blocks must only contain a single 'path'.", lineNum=startTag.line)
                 return []
         if "macros" in info:
             for k, v in info.items():
@@ -875,27 +853,23 @@ def transformInclude(lines: list[str], tagName: str, firstLine: str, lineNum: in
                 else:
                     m.die(
                         f"Include block defines the '{k}' local macro more than once.",
-                        lineNum=lineNum,
+                        lineNum=startTag.line,
                     )
                     return []
     if path:
-        el = "<pre class=include path='{}'".format(h.escapeAttr(path))
+        newStartTag.attrs["path"] = path
         for i, (macroName, macroVal) in enumerate(macros.items()):
-            el += " macro-{}='{} {}'".format(i, macroName, h.escapeAttr(macroVal))
-        el += f"{lineNumAttr}></pre>"
+            newStartTag.attrs[f"macro-{i}"] = f"{macroName} {macroVal}"
 
         indent = getWsPrefix(firstLine)
-        return [indent + el]
+        return [indent + str(newStartTag) + "</pre>"]
     return []
 
 
-def transformIncludeCode(
-    lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT
-) -> list[str]:
-    lineNumAttr = ""
-    if lineNum is not None:
-        lineNumAttr = f" line-number={lineNum}"
-    infos = parseInfoTree(lines, doc.md.indent, lineNum)
+def transformIncludeCode(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
+    newStartTag = startTag.clone(tag="pre")
+    newStartTag.classes.add("include-code")
+    infos = parseInfoTree(lines, doc.md.indent, lineNum=startTag.line)
     path = None
     highlight = None
     lineStart = None
@@ -909,7 +883,7 @@ def transformIncludeCode(
             else:
                 m.die(
                     "Include-code blocks must only contain a single 'path'.",
-                    lineNum=lineNum,
+                    lineNum=startTag.line,
                 )
         if "highlight" in info:
             if highlight is None:
@@ -917,7 +891,7 @@ def transformIncludeCode(
             else:
                 m.die(
                     "Include-code blocks must only contain a single 'highlight'.",
-                    lineNum=lineNum,
+                    lineNum=startTag.line,
                 )
         if "line-start" in info:
             if lineStart is None:
@@ -925,7 +899,7 @@ def transformIncludeCode(
             else:
                 m.die(
                     "Include-code blocks must only contain a single 'line-start'.",
-                    lineNum=lineNum,
+                    lineNum=startTag.line,
                 )
         if "show" in info:
             show.extend(info["show"])
@@ -937,29 +911,26 @@ def transformIncludeCode(
             lineNumbers = False
 
     if path:
-        attrs = lineNumAttr
-        attrs += " path='{}'".format(h.escapeAttr(path))
+        newStartTag.attrs["path"] = path
         if highlight:
-            attrs += " highlight='{}'".format(h.escapeAttr(highlight))
+            newStartTag.attrs["highlight"] = highlight
         if lineStart:
-            attrs += " line-start='{}'".format(h.escapeAttr(lineStart))
+            newStartTag.attrs["line-start"] = lineStart
         if show:
-            attrs += " data-code-show='{}'".format(h.escapeAttr(",".join(show)))
+            newStartTag.attrs["data-code-show"] = ",".join(show)
         if lineHighlight:
-            attrs += " line-highlight='{}'".format(h.escapeAttr(",".join(lineHighlight)))
+            newStartTag.attrs["line-highlight"] = ",".join(lineHighlight)
         if lineNumbers:
-            attrs += " line-numbers"
-        el = f"<pre class=include-code{attrs}></pre>"
+            newStartTag.attrs["line-numbers"] = ""
         indent = getWsPrefix(firstLine)
-        return [indent + el]
+        return [indent + str(newStartTag) + "</pre>"]
     return []
 
 
-def transformIncludeRaw(lines: list[str], tagName: str, firstLine: str, lineNum: int | None, doc: t.SpecT) -> list[str]:
-    lineNumAttr = ""
-    if lineNum is not None:
-        lineNumAttr = f" line-number={lineNum}"
-    infos = parseInfoTree(lines, doc.md.indent, lineNum)
+def transformIncludeRaw(lines: list[str], startTag: h.StartTag, firstLine: str, doc: t.SpecT) -> list[str]:
+    newStartTag = startTag.clone(tag="pre")
+    newStartTag.classes.add("include-raw")
+    infos = parseInfoTree(lines, doc.md.indent, lineNum=startTag.line)
     path = None
     for info in infos:
         if "path" in info:
@@ -968,15 +939,13 @@ def transformIncludeRaw(lines: list[str], tagName: str, firstLine: str, lineNum:
             else:
                 m.die(
                     "Include-raw blocks must only contain a single 'path'.",
-                    lineNum=lineNum,
+                    lineNum=startTag.line,
                 )
 
     if path:
-        attrs = lineNumAttr
-        attrs += " path='{}'".format(h.escapeAttr(path))
-        el = f"<pre class=include-raw{attrs}></pre>"
+        newStartTag.attrs["path"] = path
         indent = getWsPrefix(firstLine)
-        return [indent + el]
+        return [indent + str(newStartTag) + "</pre>"]
     return []
 
 
@@ -1059,189 +1028,3 @@ def parseInfoTree(lines: list[str], indent: int = 4, lineNum: int | None = 0) ->
     # Grab the last bit of data.
     extendData(datas, infoLevels[: lastIndent + 1])
     return datas
-
-
-def classesFromLine(line: Line) -> set[str]:
-    tag = parseTag(line.text.strip(), lineNumber=line.i)
-    if tag is None:
-        return set()
-    if "class" not in tag.attrs:
-        return set()
-    return set(tag.attrs["class"].strip().split())
-
-
-@attr.s(slots=True)
-class StartTag:
-    tag: str = attr.ib()
-    attrs: dict[str, str] = attr.ib(default=attr.Factory(dict))
-
-
-def parseTag(text: str, lineNumber: int | None) -> StartTag | None:
-    """
-    Parses a tag from a string,
-    conformant to the HTML parsing algorithm.
-    The text must start with the opening < character.
-    """
-
-    def parseerror(index: int, state: str) -> None:
-        m.die(
-            f"Tried to parse a start tag from '{text}', but failed at character {index} '{text[index]}' and parse-state '{state}'.",
-            lineNum=lineNumber,
-        )
-
-    def eof(i: int, text: str) -> bool:
-        return i >= len(text)
-
-    i = 0
-    state = "data"
-    while True:
-        if eof(i, text):
-            parseerror(i, state)
-            return None
-        if state == "data":
-            if text[i] == "<":
-                state = "tag-open"
-                i += 1
-                continue
-            parseerror(i, state)
-            return None
-        if state == "tag-open":
-            if text[i].isalpha():
-                state = "tag-name"
-                continue
-            parseerror(i, state)
-            return None
-        if state == "tag-name":
-            tagname = ""
-            while not eof(i, text) and re.match(r"[^\s/>]", text[i]):
-                tagname += text[i].lower()
-                i += 1
-            tag = StartTag(tagname)
-            if text[i] == ">":
-                return tag
-            if text[i] == "/":
-                state = "self-closing-start-tag"
-                i += 1
-                continue
-            if text[i].isspace():
-                state = "before-attribute-name"
-                i += 1
-                continue
-            parseerror(i, state)
-            return None
-        if state == "self-closing-start-tag":
-            if text[i] == ">":
-                return tag
-            parseerror(i, state)
-            return None
-        if state == "before-attribute-name":
-            if text[i].isspace():
-                i += 1
-                continue
-            if text[i] == "/" or text[i] == ">":
-                state = "after-attribute-name"
-                continue
-            if text[i] == "=":
-                parseerror(i, state)
-                return None
-            state = "attribute-name"
-            continue
-        if state == "attribute-name":
-            attrName = ""
-            while not eof(i, text) and re.match(r"[^\s/>=\"'<]", text[i]):
-                attrName += text[i]
-                i += 1
-            tag.attrs[attrName] = ""
-            if text[i].isspace() or text[i] == "/" or text[i] == ">":
-                state = "after-attribute-name"
-                continue
-            if text[i] == "=":
-                state = "before-attribute-value"
-                i += 1
-                continue
-            parseerror(i, state)
-            return None
-        if state == "after-attribute-name":
-            if text[i].isspace():
-                i += 1
-                continue
-            if text[i] == "/":
-                state = "self-closing-start-tag"
-                i += 1
-                continue
-            if text[i] == "=":
-                state = "before-attribute-value"
-                i += 1
-                continue
-            if text[i] == ">":
-                return tag
-            state = "attribute-name"
-            continue
-        if state == "before-attribute-value":
-            if text[i].isspace():
-                i += 1
-                continue
-            if text[i] == '"':
-                state = "attribute-value-double-quoted"
-                i += 1
-                continue
-            if text[i] == "'":
-                state = "attribute-value-single-quoted"
-                i += 1
-                continue
-            if text[i] == "=":
-                parseerror(i, state)
-                return None
-            state = "attribute-value-unquoted"
-            continue
-        if state == "attribute-value-double-quoted":
-            attrValue = ""
-            while not eof(i, text) and not text[i] == '"':
-                attrValue += text[i]
-                i += 1
-            tag.attrs[attrName] = attrValue
-            if text[i] == '"':
-                state = "after-attribute-value-quoted"
-                i += 1
-                continue
-            parseerror(i, state)
-            return None
-        if state == "attribute-value-single-quoted":
-            attrValue = ""
-            while not eof(i, text) and not text[i] == "'":
-                attrValue += text[i]
-                i += 1
-            tag.attrs[attrName] = attrValue
-            if text[i] == "'":
-                state = "after-attribute-value-quoted"
-                i += 1
-                continue
-            parseerror(i, state)
-            return None
-        if state == "attribute-value-unquoted":
-            attrValue = ""
-            while not eof(i, text) and re.match(r"[^\s<>'\"=`]", text[i]):
-                attrValue += text[i]
-                i += 1
-            tag.attrs[attrName] = attrValue
-            if text[i].isspace():
-                state = "before-attribute-name"
-                i += 1
-                continue
-            if text[i] == ">":
-                return tag
-            parseerror(i, state)
-            return None
-        if state == "after-attribute-value-quoted":
-            if text[i].isspace():
-                state = "before-attribute-name"
-                i += 1
-                continue
-            if text[i] == "/":
-                state = "self-closing-start-tag"
-                i += 1
-                continue
-            if text[i] == ">":
-                return tag
-            parseerror(i, state)
-            return None
