@@ -7,49 +7,54 @@ import io
 import os
 import re
 import typing
+import dataclasses
 from dataclasses import dataclass, field
 
+
 from .. import messages as m, t
 from . import dom
 
 
 def test() -> None:
-    nodes = []
-    with io.open(os.path.abspath("tests/abstract001.html"), "r") as fh:
-        i = 0
-        s = Stream(fh.read())
-        _, i = parseDoctype(s, i)
-        while not s.eof(i):
-            print(s.loc(i))
-            text, i = s.skip(i, "<")
-            if text:
-                nodes.append(text)
+    import json
 
-            comment, i = parseComment(s, i)
-            if comment is not Failure:
-                nodes.append(comment)
+    with io.open(os.path.abspath("test.txt"), "r") as fh:
+        vals = "\n".join(x for x in json.load(fh).values())
+        list(nodesFromHtml(vals))
 
-            startTag, i = parseStartTag(s, i)
-            if startTag is not Failure:
-                nodes.append(startTag)
-                if startTag.tag == "script":
-                    text, i = parseScriptToEnd(s, i)
-                    nodes.append(text)
-                    nodes.append(EndTag(-1, "script"))
-                elif startTag.tag == "style":
-                    text, i = parseStyleToEnd(s, i)
-                    nodes.append(text)
-                    nodes.append(EndTag(-1, "style"))
-                elif startTag.tag == "xmp":
-                    text, i = parseXmpToEnd(s, i)
-                    nodes.append(text)
-                    nodes.append(EndTag(-1, "xmp"))
 
-            endTag, i = parseEndTag(s, i)
-            if endTag is not Failure:
-                nodes.append(endTag)
-        for node in nodes:
-            print(node, end="")
+def nodesFromHtml(data: str) -> t.Generator[str | StartTag | EndTag | Comment, None, None]:
+    i = 0
+    s = Stream(data)
+    _, i = parseDoctype(s, i)
+    while not s.eof(i):
+        text, i = s.skip(i, "<")
+        if text:
+            yield text
+
+        comment, i = parseComment(s, i)
+        if comment is not Failure:
+            yield comment
+
+        startTag, i = parseStartTag(s, i)
+        if startTag is not Failure:
+            yield startTag
+            if startTag.tag == "script":
+                text, i = parseScriptToEnd(s, i)
+                yield text
+                yield EndTag(-1, "script")
+            elif startTag.tag == "style":
+                text, i = parseStyleToEnd(s, i)
+                yield text
+                yield EndTag(-1, "style")
+            elif startTag.tag == "xmp":
+                text, i = parseXmpToEnd(s, i)
+                yield text
+                yield EndTag(-1, "xmp")
+
+        endTag, i = parseEndTag(s, i)
+        if endTag is not Failure:
+            yield endTag
 
 
 #
@@ -141,10 +146,12 @@ class Result:
 class Stream:
     _chars: str
     _lineBreaks: list[int]
+    startLine: int
 
-    def __init__(self, chars: str):
+    def __init__(self, chars: str, startLine: int=1):
         self._chars = chars
         self._lineBreaks = []
+        self.startLine = startLine
         for i, char in enumerate(chars):
             if char == "\n":
                 self._lineBreaks.append(i)
@@ -161,7 +168,7 @@ class Stream:
     def line(self, index: int) -> int:
         # Zero-based line index
         lineIndex = bisect.bisect_left(self._lineBreaks, index)
-        return lineIndex + 1
+        return lineIndex + self.startLine
 
     def col(self, index: int) -> int:
         lineIndex = bisect.bisect_left(self._lineBreaks, index)
@@ -194,15 +201,26 @@ class StartTag:
     line: int
     tag: str
     attrs: dict[str, str] = field(default_factory=dict)
+    classes: set[str] = field(default_factory=set)
 
     def __str__(self) -> str:
-        start = f"<{self.tag}:{self.line}"
+        start = f"<{self.tag} line-number={self.line}"
+        attrs = ""
         if self.attrs:
-            attrs = " " + " ".join(f'{name}="{val}"' for name, val in self.attrs.items())
-        else:
-            attrs = ""
+            attrs += " " + " ".join(f'{name}="{dom.escapeAttr(val)}"' for name, val in sorted(self.attrs.items()))
+        if self.classes:
+            attrs += f' class="{" ".join(sorted(self.classes))}"'
         return start + attrs + ">"
 
+    def finalize(self) -> StartTag:
+        if "class" in self.attrs:
+            self.classes = set(self.attrs["class"].split())
+            del self.attrs["class"]
+        return self
+
+    def clone(self, **kwargs: t.Any) -> StartTag:
+        return dataclasses.replace(self, **kwargs)
+
 
 @dataclass
 class EndTag:
@@ -210,7 +228,7 @@ class EndTag:
     tag: str
 
     def __str__(self) -> str:
-        return f"</{self.tag}:{self.line}>"
+        return f"</{self.tag}>"
 
 
 @dataclass
